# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass, field
from typing import Optional, Sequence, Union

from apischema.metadata import required

# noinspection PyProtectedMember
from uai_openlabel.serializer import JsonSnakeCaseSerializableMixin

# noinspection PyProtectedMember
from uai_openlabel.transform import (
    EulerTransformData,
    Matrix4x4TransformData,
    QuaternionTransformData,
    no_default,
)

# noinspection PyProtectedMember
from uai_openlabel.types_and_constants import CoordinateSystemType, CoordinateSystemUid

__all__: list[str] = []


@dataclass
class CoordinateSystem(JsonSnakeCaseSerializableMixin):
    """
    :param children: The Sequence of children for this coordinate system
    :param parent: Each coordinate system can declare its parent coordinate system in the hierarchy
    :param pose_wrt_parent: A default or static pose of this coordinate system with respect to the declared parent.
    :param type: The type of coordinate system is defined so reading applications have a simplified view of the hierarchy.
        Suggested types:
            scene_cs, this corresponds to static coordinate systems.
            local_cs, this is a coordinate system of a rigid body, such as a vehicle, which carries with it the sensors.
            sensor_cs, a coordinate system attached to a sensor.
            custom_cs, any other coordinate system defined by the user.
    """

    parent: CoordinateSystemUid = field(
        default_factory=lambda: no_default(field="CoordinateSystem.parent"),
        metadata=required,
    )
    type: CoordinateSystemType = field(
        default_factory=lambda: no_default(field="CoordinateSystem.type"),
        metadata=required,
    )

    children: Optional[Sequence[CoordinateSystemUid]] = field(
        default=None
    )  # TODO See 7.8 first json example: child can also be what is defined in 6.4
    pose_wrt_parent: Optional[Union[Matrix4x4TransformData, QuaternionTransformData, EulerTransformData]] = field(default=None)



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass, field
from typing import Optional, TypeVar, Union

from apischema.metadata import required

# noinspection PyProtectedMember
from uai_openlabel.serializer import JsonSnakeCaseSerializableMixin

# noinspection PyProtectedMember
from uai_openlabel.types_and_constants import (
    CoordinateSystemUid,
    Number,
    Radian,
)

# noinspection PyProtectedMember
from uai_openlabel.utils import no_default

__all__: list[str] = []


@dataclass
class Matrix4x4TransformData:
    """
    Flattened list of 16 entries encoding a 4x4 homogeneous matrix to enable transform 3D column homogeneous
    vectors 4x1 using right-multiplication of matrices: X_dst = matrix_4x4 * X_src
    """

    matrix4x4: tuple[Number, ...] = field(
        default_factory=lambda: no_default(field="Matrix4x4TransformData.matrix4x4"),
        metadata=required,
    )

    def __post_init__(self) -> None:
        if len(self.matrix4x4) != 16:
            raise ValueError("matrix4x4 must be exactly 16 elements long.")


@dataclass
class QuaternionTransformData(JsonSnakeCaseSerializableMixin):
    """
    A transform can be defined with a quaternion to encode the rotation of a coordinate system
    with respect to another, and a translation.
    """

    quaternion: tuple[Number, Number, Number, Number] = field(
        default_factory=lambda: no_default(field="QuaternionTransformData.quaternion"),
        metadata=required,
    )
    translation: tuple[Number, Number, Number] = field(
        default_factory=lambda: no_default(field="QuaternionTransformData.translation"),
        metadata=required,
    )


@dataclass
class EulerTransformData(JsonSnakeCaseSerializableMixin):
    """
    A transform can be defined with a sequence of Euler angles to encode the rotation of a coordinate system
    with respect to another and a translation.
    """

    euler_angles: tuple[Radian, Radian, Radian] = field(
        default_factory=lambda: no_default(field="EulerTransformData.euler_angles"),
        metadata=required,
    )
    sequence: Optional[str] = field(default="ZYX")
    translation: tuple[Number, Number, Number] = field(
        default_factory=lambda: no_default(field="EulerTransformData.translation"),
        metadata=required,
    )


T = TypeVar("T", bound="Transform")


@dataclass
class Transform(JsonSnakeCaseSerializableMixin):
    """
    The transforms between coordinate systems may also be defined for each frame, overriding the default static pose.
    Transforms are defined with a friendly name used as index and the following properties:

    src: The name of the source coordinate system. This shall be the name of a valid (declared) coordinate system.
    dst: The destination coordinate system. This shall be the name of a valid (declared) coordinate system.
    transform_src_to_dst: This is the transform expressed in algebraic form, for example, as a 4x4 matrix enclosing a 3D rotation and a 3D translation between the coordinate systems.
    """

    dst: CoordinateSystemUid = field(default_factory=lambda: no_default(field="Transform.dst"), metadata=required)
    src: CoordinateSystemUid = field(default_factory=lambda: no_default(field="Transform.src"), metadata=required)
    transform_src_to_dst: Union[Matrix4x4TransformData, QuaternionTransformData, EulerTransformData] = field(
        default_factory=lambda: no_default(field="Transform.transform_src_to_dst"),
        metadata=required,
    )

    @classmethod
    def no_rotation_example(
        cls: type[T],
        src: CoordinateSystemUid,
        dst: CoordinateSystemUid,
        translation: tuple[float, float, float],
    ) -> T:
        # fmt: off
        translation_without_rotation = (
            1.0, 0.0, 0.0, translation[0], 
            0.0, 1.0, 0.0, translation[1], 
            0.0, 0.0, 1.0, translation[2], 
            0.0, 0.0, 0.0, 1.0
        )
        # fmt: on
        return cls(
            src=src,
            dst=dst,
            transform_src_to_dst=Matrix4x4TransformData(matrix4x4=translation_without_rotation),
        )



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# noinspection PyProtectedMember
from uai_openlabel.coordinate_system import CoordinateSystem

# noinspection PyProtectedMember
from uai_openlabel.data_types.data_pointer_types import (
    GenericDataType,
    GeometricDataType,
)

# noinspection PyProtectedMember
from uai_openlabel.data_types.data_to_pointer_type_mapping import (
    map_data_to_data_pointer_type,
)

# noinspection PyProtectedMember
from uai_openlabel.data_types.generic_data import (
    Attributes,
    BooleanData,
    BooleanType,
    GenericData,
    NumberData,
    NumberType,
    TextData,
    TextType,
    VectorData,
    VectorType,
)

# noinspection PyProtectedMember
from uai_openlabel.data_types.geometric_data import (
    GeometricData,
    Poly2D,
    Poly2DMode,
    Poly3D,
    RotatedTwoDBoundingBox,
    ThreeDBoundingBoxEuler,
    ThreeDBoundingBoxQuaternion,
    TwoDBoundingBox,
)

# noinspection PyProtectedMember
from uai_openlabel.elements.action import Action, ActionInFrame

# noinspection PyProtectedMember
from uai_openlabel.elements.context import Context, ContextInFrame

# noinspection PyProtectedMember
from uai_openlabel.elements.element_data_pointer import (
    ElementDataPointer,
)

# noinspection PyProtectedMember
from uai_openlabel.elements.event import Event, EventInFrame

# noinspection PyProtectedMember
from uai_openlabel.elements.object import Object, ObjectData, ObjectInFrame

# noinspection PyProtectedMember
from uai_openlabel.elements.relation import RdfAgent, RdfAgentType, Relation

# noinspection PyProtectedMember
from uai_openlabel.frame import Frame, FrameProperties

# noinspection PyProtectedMember
from uai_openlabel.frame_interval import FrameInterval

# noinspection PyProtectedMember
from uai_openlabel.metadata import Metadata

# noinspection PyProtectedMember
from uai_openlabel.ontology import DetailedOntology

# noinspection PyProtectedMember
from uai_openlabel.openlabel import OpenLabel

# noinspection PyProtectedMember
from uai_openlabel.stream.camera_intrinsics import (
    CustomCameraIntrinsics,
    FisheyeCameraIntrinsics,
    PinholeCameraIntrinsics,
)

# noinspection PyProtectedMember
from uai_openlabel.stream.stream import Stream, StreamType

# noinspection PyProtectedMember
from uai_openlabel.stream.stream_properties import (
    CustomCameraStreamProperties,
    FisheyeCameraStreamProperties,
    PinholeCameraStreamProperties,
    StreamProperties,
    SyncByFrameShift,
    SyncByFrameStream,
)

# noinspection PyProtectedMember
from uai_openlabel.tag import Tag

# noinspection PyProtectedMember
from uai_openlabel.transform import (
    EulerTransformData,
    Matrix4x4TransformData,
    QuaternionTransformData,
    Transform,
)

# noinspection PyProtectedMember
from uai_openlabel.types_and_constants import (
    URI,
    ActionUid,
    AttributeName,
    ContextUid,
    CoordinateSystemType,
    CoordinateSystemUid,
    ElementUid,
    EventUid,
    FrameUid,
    Meter,
    Number,
    ObjectUid,
    OntologyUid,
    Pixel,
    Radian,
    RelationUid,
    ResourceUid,
    StreamUid,
    TagUid,
    Uid,
)

# noinspection PyProtectedMember
from uai_openlabel.utils import convert_values, no_default

__all__ = [
    # data_types
    "Attributes",
    "BooleanType",
    "BooleanData",
    "NumberType",
    "NumberData",
    "TextType",
    "TextData",
    "VectorType",
    "VectorData",
    "GenericData",
    "TwoDBoundingBox",
    "RotatedTwoDBoundingBox",
    "ThreeDBoundingBoxQuaternion",
    "ThreeDBoundingBoxEuler",
    "Poly2D",
    "Poly2DMode",
    "Poly3D",
    "GeometricData",
    # elements
    "Action",
    "ActionInFrame",
    "Context",
    "ContextInFrame",
    "GeometricDataType",
    "GenericDataType",
    "ElementDataPointer",
    "map_data_to_data_pointer_type",
    "Event",
    "EventInFrame",
    "ObjectData",
    "Object",
    "ObjectInFrame",
    "RdfAgentType",
    "RdfAgent",
    "Relation",
    # stream
    "PinholeCameraIntrinsics",
    "FisheyeCameraIntrinsics",
    "CustomCameraIntrinsics",
    "Stream",
    "StreamType",
    "SyncByFrameStream",
    "SyncByFrameShift",
    "PinholeCameraStreamProperties",
    "FisheyeCameraStreamProperties",
    "CustomCameraStreamProperties",
    "StreamProperties",
    # types_and_constants
    "URI",
    "Uid",
    "FrameUid",
    "ObjectUid",
    "EventUid",
    "OntologyUid",
    "ResourceUid",
    "TagUid",
    "StreamUid",
    "ElementUid",
    "ActionUid",
    "ContextUid",
    "RelationUid",
    "AttributeName",
    "CoordinateSystemUid",
    "CoordinateSystemType",
    "Number",
    "Meter",
    "Radian",
    "Pixel",
    # All the rest
    "CoordinateSystem",
    "FrameInterval",
    "FrameProperties",
    "Frame",
    "Metadata",
    "DetailedOntology",
    "OpenLabel",
    "Tag",
    "Matrix4x4TransformData",
    "QuaternionTransformData",
    "EulerTransformData",
    "Transform",
    "no_default",
    "convert_values",
]



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from typing import Union

__all__: list[str] = []

# noinspection PyProtectedMember
from uai_openlabel.data_types.data_pointer_types import (
    GenericDataType,
    GeometricDataType,
)

# noinspection PyProtectedMember
from uai_openlabel.data_types.generic_data import (
    BooleanData,
    GenericData,
    NumberData,
    TextData,
    VectorData,
)

# noinspection PyProtectedMember
from uai_openlabel.data_types.geometric_data import (
    GeometricData,
    Poly2D,
    Poly3D,
    RotatedTwoDBoundingBox,
    ThreeDBoundingBoxEuler,
    ThreeDBoundingBoxQuaternion,
    TwoDBoundingBox,
)

# Don't export this, only the function
DATA_TO_POINTER_TYPE_MAPPING: dict[type[Union[GenericData, GeometricData]], Union[GenericDataType, GeometricDataType]] = {
    # GenericData
    BooleanData: GenericDataType.Boolean,
    NumberData: GenericDataType.Number,
    TextData: GenericDataType.Text,
    VectorData: GenericDataType.Vector,
    # GeometricData
    TwoDBoundingBox: GeometricDataType.TwoDBoundingBox,
    RotatedTwoDBoundingBox: GeometricDataType.RotatedTwoDBoundingBox,
    ThreeDBoundingBoxEuler: GeometricDataType.Cuboid,
    ThreeDBoundingBoxQuaternion: GeometricDataType.Cuboid,
    Poly2D: GeometricDataType.TwoDPolygon,
    Poly3D: GeometricDataType.ThreeDPolygon,
}


def map_data_to_data_pointer_type(data: Union[GenericData, GeometricData]) -> Union[GenericDataType, GeometricDataType]:
    if data.__class__ not in DATA_TO_POINTER_TYPE_MAPPING:
        raise KeyError(data)
    return DATA_TO_POINTER_TYPE_MAPPING[data.__class__]



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass, field
from typing import Mapping, Optional, Sequence, Union

from apischema.metadata import required

# noinspection PyProtectedMember
from uai_openlabel.data_types.data_pointer_types import (
    GenericDataType,
    GeometricDataType,
)

# noinspection PyProtectedMember
from uai_openlabel.frame_interval import FrameInterval, no_default

# noinspection PyProtectedMember
from uai_openlabel.serializer import JsonSnakeCaseSerializableMixin

# noinspection PyProtectedMember
from uai_openlabel.types_and_constants import AttributeName

__all__: list[str] = []


@dataclass
class ElementDataPointer(JsonSnakeCaseSerializableMixin):
    frame_intervals: Sequence[FrameInterval] = field(
        default_factory=lambda: no_default(field="ElementDataPointer.frame_intervals"),
        metadata=required,
    )
    """List of frame intervals of the element data pointed by this pointer."""

    attribute_pointers: Optional[Mapping[AttributeName, GenericDataType]] = field(default=None)
    """
    This is a JSON object which contains pointers to the attributes of the element data pointed by this pointer. 
    The attributes pointer keys shall be the \"name\" of the attribute of the element data this pointer points to.
    """

    type: Optional[Union[GenericDataType, GeometricDataType]] = field(default=None)



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from enum import Enum

__all__: list[str] = []


class GenericDataType(Enum):
    Boolean = "boolean"
    Number = "num"
    Text = "text"
    Vector = "vec"


class GeometricDataType(Enum):
    TwoDBoundingBox = "bbox"
    RotatedTwoDBoundingBox = "rbbox"
    TwoDPolygon = "poly2d"
    ThreeDPolygon = "poly3d"
    Cuboid = "cuboid"
    Image = "image"
    Matrix = "mat"
    BinaryPayload = "binary"
    Point2d = "point2d"
    Point3d = "point3d"
    LineReference = "line_reference"
    AreaReference = "area_reference"
    Mesh = "mesh"



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Sequence

from apischema.metadata import required

# noinspection PyProtectedMember
from uai_openlabel.frame_interval import FrameInterval, no_default

# noinspection PyProtectedMember
from uai_openlabel.serializer import JsonSnakeCaseSerializableMixin

# noinspection PyProtectedMember
from uai_openlabel.types_and_constants import ElementUid, OntologyUid, ResourceUid

__all__: list[str] = []


class RdfAgentType(Enum):
    Object = "object"
    Action = "action"
    Event = "event"
    Context = "context"


@dataclass
class RdfAgent(JsonSnakeCaseSerializableMixin):
    """
    :param type: The OpenLABEL type of element.
    :param uid: The element UID this RDF agent refers to.
    """

    type: RdfAgentType = field(default_factory=lambda: no_default(field="RdfAgent.type"), metadata=required)
    uid: ElementUid = field(default_factory=lambda: no_default(field="RdfAgent.uid"), metadata=required)


@dataclass
class Relation(JsonSnakeCaseSerializableMixin):
    """
    A relation is a type of element which connects two or more other elements, for example, objects, actions, contexts, or events.
    RDF triples are used to structure the connection with one or more subjects, a predicate, and one or more semantic objects.

    :param frame_intervals: The array of frame intervals where this relation exists or is defined.
    :param name: Name of the relation. It is a friendly name and not used for indexing.
    :param ontology_uid: This is the UID of the ontology where the type of this relation is defined.
    :param rdf_objects: This is the Sequence of RDF semantic objects of this relation.
    :param rdf_subjects: This is the Sequence of RDF semantic subjects of this relation.
    :param resource_uid: This is a JSON object that contains links to external resources.
        Resource_uid keys are strings containing numerical UIDs or 32 bytes UUIDs.
        Resource_uid values are strings describing the identifier of the element in the external resource.
    :param type: The type of a relation defines the class the predicated of the relation corresponds to.
    """

    name: str = field(default_factory=lambda: no_default(field="Relation.name"), metadata=required)
    type: str = field(default_factory=lambda: no_default(field="Relation.type"), metadata=required)
    rdf_objects: Sequence[RdfAgent] = field(
        default_factory=lambda: no_default(field="Relation.rdf_objects"),
        metadata=required,
    )
    rdf_subjects: Sequence[RdfAgent] = field(
        default_factory=lambda: no_default(field="Relation.rdf_subjects"),
        metadata=required,
    )

    frame_intervals: Optional[Sequence[FrameInterval]] = field(default=None)
    ontology_uid: Optional[OntologyUid] = field(default=None)
    resource_uid: Optional[ResourceUid] = field(default=None)



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Sequence, Union

from apischema.metadata import required

# noinspection PyProtectedMember
from uai_openlabel.data_types.generic_data import Attributes

# noinspection PyProtectedMember
from uai_openlabel.serializer import JsonSnakeCaseSerializableMixin

# noinspection PyProtectedMember
from uai_openlabel.types_and_constants import (
    AttributeName,
    CoordinateSystemUid,
    Number,
    Pixel,
    Radian,
)

# noinspection PyProtectedMember
from uai_openlabel.utils import convert_values, no_default

__all__: list[str] = []


@dataclass
class TwoDBoundingBox(JsonSnakeCaseSerializableMixin):
    """
    A 2D bounding box is defined as a rectangle by an array of four floating point numbers:

    Attribute	Unit	Description
    x           px      Specify the x-coordinate of the center of the rectangle.
    y           px      Specify the y-coordinate of the center of the rectangle.
    w           px      Specify the width of the rectangle in the x/y-coordinate system (horizontal, x-coordinate dimension).
    h           px      Specify the height of the rectangle in the x/y-coordinate system (vertical, y-coordinate dimension).
    """

    val: tuple[Pixel, Pixel, Pixel, Pixel] = field(
        default_factory=lambda: no_default(field="TwoDBoundingBox.val"),
        metadata=required,
    )

    attributes: Optional[Attributes] = field(default=None)
    coordinate_system: Optional[CoordinateSystemUid] = field(default=None)
    name: AttributeName = field(
        default_factory=lambda: no_default(field="TwoDBoundingBox.name"),
        metadata=required,
    )

    def __post_init__(self) -> None:
        """apischema doesn't allow subclasses of the annotated types, so we need to cast them."""
        field_name_for_logging = f"{self.__class__.__name__}.val"
        converted_val = convert_values(
            values=self.val,
            conversion_target=float,
            dont_convert=[int],
            field_name_for_logging=field_name_for_logging,
        )
        if len(converted_val) != 4:
            raise ValueError(f"{field_name_for_logging} must have a length of 4")
        self.val = tuple(converted_val)  # type: ignore


@dataclass
class RotatedTwoDBoundingBox(JsonSnakeCaseSerializableMixin):
    """
    The value of a 2D rotated bounding box is defined as a 5-dimensional vector by five numbers:

    Attribute	Unit	Description
    x           px      Specify the x-coordinate of the center of the rectangle.
    y           px      Specify the y-coordinate of the center of the rectangle.
    w           px      Specify the width of the rectangle in the x/y-coordinate system (horizontal, x-coordinate dimension).
    h           px      Specify the height of the rectangle in the x/y-coordinate system (vertical, y-coordinate dimension).
    alpha       rad     Specifies the rotation of the rotated bounding box. It is defined as a right-handed rotation,
                        meaning positive from x-axes to y-axes. The origin of rotation is placed at the center of the bounding box, meaning x, y.
    """

    val: tuple[Pixel, Pixel, Pixel, Pixel, Radian] = field(
        default_factory=lambda: no_default(field="RotatedTwoDBoundingBox.val"),
        metadata=required,
    )

    attributes: Optional[Attributes] = field(default=None)
    coordinate_system: Optional[CoordinateSystemUid] = field(default=None)
    name: AttributeName = field(
        default_factory=lambda: no_default(field="RotatedTwoDBoundingBox.name"),
        metadata=required,
    )

    def __post_init__(self) -> None:
        """apischema doesn't allow subclasses of the annotated types, so we need to cast them."""
        field_name_for_logging = f"{self.__class__.__name__}.val"
        converted_val = convert_values(
            values=self.val,
            conversion_target=float,
            dont_convert=[int],
            field_name_for_logging=field_name_for_logging,
        )
        if len(converted_val) != 5:
            raise ValueError(f"{field_name_for_logging} must have a length of 5")
        self.val = tuple(converted_val)  # type: ignore


@dataclass
class ThreeDBoundingBoxQuaternion(JsonSnakeCaseSerializableMixin):
    """
    A cuboid in 3D Euclidean space that is defined as val=(x, y, z, qa, qb, qc, qd, sx, sy, and sz), where:

    Attribute	unit	Description
    x           m       The x-coordinate of the 3D position of the center of the cuboid.
    y           m       The y-coordinate of the 3D position of the center of the cuboid.
    z           m       The z-coordinate of the 3D position of the center of the cuboid.
    qa                  Quaternion in non-unit form (x, y, z, and w) as in the SciPy convention.
    qb                  Quaternion in non-unit form (x, y, z, and w) as in the SciPy convention.
    qc                  Quaternion in non-unit form (x, y, z, and w) as in the SciPy convention.
    qd                  Quaternion in non-unit form (x, y, z, and w) as in the SciPy convention.
    sx          m       The x-dimension of the cuboid or the x-coordinate.
    sy          m       The y-dimension of the cuboid or the y-coordinate.
    sz          m       The z-dimension of the cuboid or the z-coordinate.
    """

    val: tuple[Number, Number, Number, Number, Number, Number, Number, Number, Number, Number] = field(
        default_factory=lambda: no_default(field="ThreeDBoundingBoxQuaternion.val"),
        metadata=required,
    )

    attributes: Optional[Attributes] = field(default=None)
    coordinate_system: Optional[CoordinateSystemUid] = field(default=None)
    name: AttributeName = field(
        default_factory=lambda: no_default(field="ThreeDBoundingBoxQuaternion.name"),
        metadata=required,
    )

    def __post_init__(self) -> None:
        """apischema doesn't allow subclasses of the annotated types, so we need to cast them."""
        field_name_for_logging = f"{self.__class__.__name__}.val"
        converted_val = convert_values(
            values=self.val,
            conversion_target=float,
            dont_convert=[int],
            field_name_for_logging=field_name_for_logging,
        )
        if len(converted_val) != 10:
            raise ValueError(f"{field_name_for_logging} must have a length of 10")
        self.val = tuple(converted_val)  # type: ignore


@dataclass
class ThreeDBoundingBoxEuler(JsonSnakeCaseSerializableMixin):
    """
    A cuboid in 3D Euclidean space defined as val=(x, y, z, rx, ry, rz, sx, sy, and sz), where:

    Attribute	unit	Description
    x           m       The x-coordinate of the 3D position of the center of the cuboid.
    y           m       The y-coordinate of the 3D position of the center of the cuboid.
    z           m       The z-coordinate of the 3D position of the center of the cuboid.
    rx          rad     Euler angles, rx = roll.
    ry          rad     Euler angles, ry = pitch.
    rz          rad     Euler angles, rz = yaw.
    sx          m       The x-dimension of the cuboid or the x-coordinate.
    sy          m       The y-dimension of the cuboid or the y-coordinate.
    sz          m       The z-dimension of the cuboid or the z-coordinate.
    """

    val: tuple[Number, Number, Number, Number, Number, Number, Number, Number, Number] = field(
        default_factory=lambda: no_default(field="ThreeDBoundingBoxEuler.val"),
        metadata=required,
    )

    attributes: Optional[Attributes] = field(default=None)
    coordinate_system: Optional[CoordinateSystemUid] = field(default=None)
    name: AttributeName = field(
        default_factory=lambda: no_default(field="ThreeDBoundingBoxEuler.name"),
        metadata=required,
    )

    def __post_init__(self) -> None:
        """apischema doesn't allow subclasses of the annotated types, so we need to cast them."""
        field_name_for_logging = f"{self.__class__.__name__}.val"
        converted_val = convert_values(
            values=self.val,
            conversion_target=float,
            dont_convert=[int],
            field_name_for_logging=field_name_for_logging,
        )
        if len(converted_val) != 9:
            raise ValueError(f"{ field_name_for_logging} must have a length of 9")
        self.val = tuple(converted_val)  # type: ignore


class Poly2DMode(Enum):
    """
    Mode of the polyline list of values:
     - "MODE_POLY2D_ABSOLUTE" determines that the poly2d list contains the sequence of (x, y) values of all points of the polyline.
     - "MODE_POLY2D_RELATIVE" specifies that only the first point of the sequence is defined with its (x, y) values, while all the rest are defined relative to it.
     - "MODE_POLY2D_SRF6DCC" specifies that SRF6DCC chain code method is used.
     - "MODE_POLY2D_RS6FCC" specifies that the RS6FCC method is used.
    """

    Absolute = "MODE_POLY2D_ABSOLUTE"
    Relative = "MODE_POLY2D_RELATIVE"
    SRF6DCC = "MODE_POLY2D_SRF6DCC"
    RS6FCC = "MODE_POLY2D_RS6FCC"


@dataclass
class Poly2D(JsonSnakeCaseSerializableMixin):
    """
    A 2D polyline defined as a sequence of 2D points.
    """

    attributes: Optional[Attributes] = field(default=None)

    closed: bool = field(default_factory=lambda: no_default(field="Poly2D.closed"), metadata=required)
    """
    A boolean that defines whether the polyline is closed or not. 
    In case it is closed, it is assumed that the last point of the sequence is connected with the first one.
    """

    coordinate_system: Optional[CoordinateSystemUid] = field(default=None)
    """Name of the coordinate system in respect of which this object data is expressed."""

    hierarchy: Optional[Sequence[int]] = field(default=None)
    """Hierarchy of the 2D polyline in the context of a set of 2D polylines."""

    mode: Poly2DMode = field(default_factory=lambda: no_default(field="Poly2D.mode"), metadata=required)

    name: AttributeName = field(default_factory=lambda: no_default(field="Poly2D.name"), metadata=required)
    """This is a string encoding the name of this object data. It is used as index inside the corresponding object data pointers."""

    val: Union[Sequence[Number], Sequence[str]] = field(
        default_factory=lambda: no_default(field="Poly2D.val"), metadata=required
    )
    """List of numerical values of the polyline, according to its mode."""

    def __post_init__(self) -> None:
        if self.hierarchy is not None and len(self.hierarchy) != 4:
            raise ValueError("Poly2d.hierarchy must be of length 4")

        field_name_for_logging = f"{self.__class__.__name__}.val"
        converted_val = convert_values(
            values=self.val,
            conversion_target=float,
            dont_convert=[int, str],
            field_name_for_logging=field_name_for_logging,
        )
        self.val = tuple(converted_val)


@dataclass
class Poly3D(JsonSnakeCaseSerializableMixin):
    """
    A 3D polyline defined as a sequence of 3D points.
    """

    attributes: Optional[Attributes] = field(default=None)

    closed: bool = field(default_factory=lambda: no_default(field="Poly3D.closed"), metadata=required)
    """
    A boolean that defines whether the polyline is closed or not. 
    In case it is closed, it is assumed that the last point of the sequence is connected with the first one.
    """

    coordinate_system: Optional[CoordinateSystemUid] = field(default=None)
    """Name of the coordinate system in respect of which this object data is expressed."""

    name: AttributeName = field(default_factory=lambda: no_default(field="Poly3D.name"), metadata=required)
    """This is a string encoding the name of this object data. It is used as index inside the corresponding object data pointers."""

    val: Sequence[Number] = field(default_factory=lambda: no_default(field="Poly3D.val"), metadata=required)
    """List of numerical values of the polyline."""

    def __post_init__(self) -> None:
        """apischema doesn't allow subclasses of the annotated types, so we need to cast them."""
        field_name_for_logging = f"{self.__class__.__name__}.val"
        converted_val = convert_values(
            values=self.val,
            conversion_target=float,
            dont_convert=[int],
            field_name_for_logging=field_name_for_logging,
        )
        self.val = tuple(converted_val)


GeometricData = Union[
    TwoDBoundingBox,
    RotatedTwoDBoundingBox,
    ThreeDBoundingBoxEuler,
    ThreeDBoundingBoxQuaternion,
    Poly2D,
    Poly3D,
]



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass, field
from typing import Any, Mapping, Optional, Sequence, TypeVar, Union, cast

from uai_openlabel import ElementDataPointer, map_data_to_data_pointer_type

# noinspection PyProtectedMember
from uai_openlabel.coordinate_system import CoordinateSystem

# noinspection PyProtectedMember
from uai_openlabel.elements.action import Action

# noinspection PyProtectedMember
from uai_openlabel.elements.context import Context

# noinspection PyProtectedMember
from uai_openlabel.elements.event import Event

# noinspection PyProtectedMember
from uai_openlabel.elements.object import Object, ObjectData, ObjectInFrame

# noinspection PyProtectedMember
from uai_openlabel.elements.relation import Relation

# noinspection PyProtectedMember
from uai_openlabel.frame import Frame

# noinspection PyProtectedMember
from uai_openlabel.frame_interval import FrameInterval

# noinspection PyProtectedMember
from uai_openlabel.metadata import Metadata

# noinspection PyProtectedMember
from uai_openlabel.ontology import DetailedOntology

# noinspection PyProtectedMember
from uai_openlabel.serializer import JsonSnakeCaseSerializableMixin

# noinspection PyProtectedMember
from uai_openlabel.stream.stream import Stream

# noinspection PyProtectedMember
from uai_openlabel.tag import Tag

# noinspection PyProtectedMember
from uai_openlabel.types_and_constants import (
    URI,
    ActionUid,
    AttributeName,
    ContextUid,
    CoordinateSystemUid,
    EventUid,
    ObjectUid,
    OntologyUid,
    RelationUid,
    ResourceUid,
    StreamUid,
    TagUid,
    Uid,
)

__all__: list[str] = []

T = TypeVar("T", bound="OpenLabel")
J = TypeVar("J", bound=JsonSnakeCaseSerializableMixin)


@dataclass
class OpenLabel(JsonSnakeCaseSerializableMixin):
    """
    The version of the schema shall be defined inside the metadata structure, using the key schema_version.
    All other entries are optional.

    :param actions: This is the JSON object of OpenLABEL actions. AveasAction keys are strings containing numerical UIDs or 32 bytes UUIDs.
    :param contexts: This is the JSON object of OpenLABEL contexts. Context keys are strings containing numerical UIDs or 32 bytes UUIDs.
    :param coordinate_systems: This is a JSON object which contains OpenLABEL coordinate systems.
        Coordinate system keys can be any string, for example, a friendly coordinate system name.
    :param events: This is the JSON object of OpenLABEL events. Event keys are strings containing numerical UIDs or 32 bytes UUIDs.
    :param frame_intervals: This is an array of frame intervals.
    :param frames: This is the JSON object of frames that contain the dynamic, timewise, annotations.
        Keys are strings containing numerical frame identifiers, which are denoted as master frame numbers.
    :param metadata: This JSON object contains information, that is, metadata, about the annotation file itself.
    :param objects: This is the JSON object of OpenLABEL objects. Object keys are strings containing numerical UIDs or 32 bytes UUIDs.
    :param ontologies: This is the JSON object of OpenLABEL ontologies. Ontology keys are strings containing numerical UIDs or 32 bytes UUIDs.
        Ontology values may be strings, for example, encoding a URI. JSON objects containing a URI string and optional Sequences of included and excluded terms.
    :param relations: This is the JSON object of OpenLABEL relations. Relation keys are strings containing numerical UIDs or 32 bytes UUIDs.
    :param resources: This is the JSON object of OpenLABEL resources. Resource keys are strings containing numerical UIDs or 32 bytes UUIDs.
        Resource values are strings that describe an external resource, for example, file name, URLs, that may be
        used to link data of the OpenLABEL annotation content with external existing content.
    :param streams: This is a JSON object which contains OpenLABEL streams. Stream keys can be any string, for example, a friendly stream name.
    :param tags: This is the JSON object of tags. Tag keys are strings containing numerical UIDs or 32 bytes UUIDs.
    """

    metadata: Metadata = field(default_factory=Metadata)

    actions: Optional[Mapping[ActionUid, Action]] = field(default=None)
    contexts: Optional[Mapping[ContextUid, Context]] = field(default=None)
    coordinate_systems: Optional[Mapping[CoordinateSystemUid, CoordinateSystem]] = field(default=None)
    events: Optional[Mapping[EventUid, Event]] = field(default=None)
    frame_intervals: Optional[Sequence[FrameInterval]] = field(default=None)
    frames: Optional[Mapping[Uid, Frame]] = field(default=None)
    objects: Optional[Mapping[ObjectUid, Object]] = field(default=None)
    ontologies: Optional[Mapping[OntologyUid, Union[URI, DetailedOntology]]] = field(default=None)
    relations: Optional[Mapping[RelationUid, Relation]] = field(default=None)
    resources: Optional[Mapping[ResourceUid, URI]] = field(default=None)
    streams: Optional[Mapping[StreamUid, Stream]] = field(default=None)
    tags: Optional[Mapping[TagUid, Tag]] = field(default=None)

    @classmethod
    def example(cls: type[T]) -> T:
        """This method creates an example OpenLabel containing the fields that we would usually also fill in our exports."""

        frames = {
            Uid("001"): Frame.example(
                timestamp="1700126000005000",  # ms
                transform_from_and_to=("reference", "world"),
                transform_translation=(1.0, 0.0, 0.0),
                frame_objects={
                    Uid("1"): ObjectInFrame.example(toggle_attribute_values=True, cuboid_translation=(1.0, 0.0, 0.0)),
                    Uid("2"): ObjectInFrame.example(toggle_attribute_values=True, cuboid_translation=(5.0, 1.0, 0.0)),
                    Uid("3"): ObjectInFrame.example(
                        toggle_attribute_values=True,
                        cuboid_translation=(10.0, 1.0, 0.0),
                    ),
                },
            ),
            Uid("002"): Frame.example(
                timestamp="1700126000005010",  # 10ms later
                transform_from_and_to=("reference", "world"),
                transform_translation=(1.1, 0.0, 0.0),
                frame_objects={
                    Uid("1"): ObjectInFrame.example(toggle_attribute_values=True, cuboid_translation=(1.1, 0.0, 0.0)),
                    Uid("2"): ObjectInFrame.example(toggle_attribute_values=True, cuboid_translation=(5.1, 1.0, 0.0)),
                    Uid("3"): ObjectInFrame.example(
                        toggle_attribute_values=True,
                        cuboid_translation=(10.1, 1.0, 0.0),
                    ),
                },
            ),
            Uid("003"): Frame.example(
                timestamp="1700126000005020",  # 10ms later
                transform_from_and_to=("reference", "world"),
                transform_translation=(1.2, 0.0, 0.0),
                frame_objects={
                    Uid("1"): ObjectInFrame.example(toggle_attribute_values=True, cuboid_translation=(1.2, 0.0, 0.0)),
                    Uid("2"): ObjectInFrame.example(toggle_attribute_values=True, cuboid_translation=(5.1, 1.0, 0.0)),
                    Uid("3"): ObjectInFrame.example(
                        toggle_attribute_values=True,
                        cuboid_translation=(10.2, 1.0, 0.0),
                    ),
                },
            ),
        }

        frame_interval = FrameInterval(frame_start=Uid("1"), frame_end=Uid("3"))

        object_data_pointers: dict[AttributeName, ElementDataPointer] = {}
        attributes_of_dynamic_objects = ObjectData.cuboid_with_dynamic_attributes_example(False, (0.0, 0.0, 0.0))
        for data in attributes_of_dynamic_objects:
            assert data.name is not None, "All attributes must have a name in this example"
            object_data_pointers[data.name] = ElementDataPointer(
                frame_intervals=[frame_interval],
                type=map_data_to_data_pointer_type(data),
            )

        objects = {
            Uid("1"): Object.car_example(
                name="ego_vehicle",
                frame_interval=frame_interval,
                object_data_pointers=object_data_pointers,
            ),
            Uid("2"): Object.car_example(
                name="car1",
                frame_interval=frame_interval,
                object_data_pointers=object_data_pointers,
            ),
            Uid("3"): Object.car_example(
                name="car2",
                frame_interval=frame_interval,
                object_data_pointers=object_data_pointers,
            ),
        }

        coordinate_systems_dict = {
            "world": CoordinateSystem(parent="", type="scene_cs", children=["reference"]),
            "reference": CoordinateSystem(parent="world", type="sensor_cs"),
        }
        return cls(
            metadata=Metadata.example(),
            actions=None,
            contexts=None,
            coordinate_systems=coordinate_systems_dict,
            events=None,
            frame_intervals=[frame_interval],
            frames=cast(Mapping, frames),
            objects=objects,
            ontologies=None,
            relations=None,
            resources=None,
            streams={"reference": Stream.example_lidar_stream("reference")},
            tags=None,
        )

    @classmethod
    def from_dict(cls: type[J], kvs: dict[str, Any], *, infer_missing: bool = False) -> J:
        """
        Any ASAM OpenLABEL JSON data shall have a root key named openlabel.
        """
        if list(kvs.keys()) == ["openlabel"]:
            kvs = kvs["openlabel"]
        return super().from_dict(kvs, infer_missing=infer_missing)

    def to_dict(
        self,
        encode_json: bool = False,
        exclude_none: bool = False,
        exclude_defaults: bool = False,
    ) -> dict:
        serialized = super().to_dict(encode_json, exclude_none, exclude_defaults)
        with_root_key = {"openlabel": serialized}
        return with_root_key



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import re
from typing import Union

__all__: list[str] = []


URI = str


UID_PATTERN = re.compile("^(-?[0-9]+|[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12})$")


class Uid(str):
    def __init__(self, val: str):
        matches_pattern = bool(UID_PATTERN.match(val))
        if not matches_pattern:
            raise ValueError(f"{val} doesn't match the OpenLABEL UID pattern {UID_PATTERN.pattern}")
        super().__init__()


# Types found in OpenLABEL root
FrameUid = Union[Uid, int]
OntologyUid = Uid
ResourceUid = Uid
TagUid = Uid
StreamUid = str


# Element UIDs
ElementUid = Uid
ActionUid = ElementUid
ContextUid = ElementUid
EventUid = ElementUid
ObjectUid = ElementUid
RelationUid = ElementUid


# Types found in element attributes
AttributeName = str
CoordinateSystemUid = str
CoordinateSystemType = str


# Units of length and rotation
Number = Union[int, float]  # Sub-pixel precision is supported

Meter = Number
Radian = Number
Pixel = Number



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import builtins
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, TypeVar

# noinspection PyProtectedMember
from uai_openlabel.serializer import JsonSnakeCaseSerializableMixin

# noinspection PyProtectedMember
from uai_openlabel.stream.stream_properties import (
    StreamProperties,
)

# noinspection PyProtectedMember
from uai_openlabel.types_and_constants import URI

# noinspection PyProtectedMember

__all__: list[str] = []


class StreamType(Enum):
    Camera = "camera"
    Lidar = "lidar"
    Radar = "radar"
    GPS_IMU = "gps_imu"
    Other = "other"


T = TypeVar("T", bound="Stream")


@dataclass
class Stream(JsonSnakeCaseSerializableMixin):
    """
    From 7.5.3 of the openLABEL standard:
    The sync field within stream_properties defines the frame number of the stream that corresponds to this frame,
    along with timestamping information, if needed.

    This feature is useful for annotating multiple cameras which might not be perfectly aligned.
    In such cases, frame 0 of the ASAM OpenLABEL JSON data corresponds to frame 0 of the first stream to occur.
    In this way, frame_stream shall identify which frame of this stream corresponds to the frame in which it is enclosed.
    """

    description: Optional[str] = field(default=None)
    type: Optional[StreamType] = field(default=None)

    stream_properties: Optional[StreamProperties] = field(default=None)
    uri: Optional[URI] = field(default=None)

    @classmethod
    def example_lidar_stream(cls: builtins.type[T], description: str) -> T:
        return cls(description=description, type=StreamType.Lidar)



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass, field
from typing import Optional, Sequence

from apischema.metadata import required

# noinspection PyProtectedMember
from uai_openlabel.serializer import JsonSnakeCaseSerializableMixin

# noinspection PyProtectedMember
from uai_openlabel.types_and_constants import Number, Pixel

# noinspection PyProtectedMember
from uai_openlabel.utils import no_default

__all__: list[str] = []


@dataclass
class FisheyeCameraIntrinsics(JsonSnakeCaseSerializableMixin):
    height_px: Pixel = field(
        default_factory=lambda: no_default(field="FisheyeCameraIntrinsics.height_px"),
        metadata=required,
    )
    width_px: Pixel = field(
        default_factory=lambda: no_default(field="FisheyeCameraIntrinsics.width_px"),
        metadata=required,
    )
    lens_coeffs: Sequence[Number] = field(
        default_factory=lambda: no_default(field="FisheyeCameraIntrinsics.lens_coeffs"),
        metadata=required,
    )

    center_x_px: Optional[Pixel] = field(default=None)
    center_y_px: Optional[Pixel] = field(default=None)
    focal_length_x: Optional[Pixel] = field(default=None)
    focal_length_y: Optional[Pixel] = field(default=None)

    def __post_init__(self) -> None:
        nr_lens_coeffs = len(self.lens_coeffs)
        if nr_lens_coeffs < 4 or nr_lens_coeffs > 5:
            raise ValueError("lens_coeffs must contain 4 or 5 elements")


@dataclass
class PinholeCameraIntrinsics(JsonSnakeCaseSerializableMixin):
    camera_matrix: Sequence[Pixel] = field(
        default_factory=lambda: no_default(field="PinholeCameraIntrinsics.camera_matrix"),
        metadata=required,
    )
    distortion_coeffs: Sequence[Pixel] = field(
        default_factory=lambda: no_default(field="PinholeCameraIntrinsics.distortion_coeffs"),
        metadata=required,
    )
    height_px: Pixel = field(
        default_factory=lambda: no_default(field="PinholeCameraIntrinsics.height_px"),
        metadata=required,
    )
    width_px: Pixel = field(
        default_factory=lambda: no_default(field="PinholeCameraIntrinsics.width_px"),
        metadata=required,
    )

    def __post_init__(self) -> None:
        if len(self.camera_matrix) != 12:
            raise ValueError("camera_matrix must consist of exactly 12 elements")

        nr_distortion_coeffs = len(self.distortion_coeffs)
        if nr_distortion_coeffs < 5 or nr_distortion_coeffs > 14:
            raise ValueError("distortion_coeffs must be between 5 and 14 elements long")


@dataclass
class CustomCameraIntrinsics(JsonSnakeCaseSerializableMixin):
    ...



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import is_dataclass
from typing import Any, Protocol, TypeVar, cast, runtime_checkable

import apischema

__all__: list[str] = []

T = TypeVar("T", bound="JsonSnakeCaseSerializableMixin")


@runtime_checkable
class DataclassLike(Protocol):
    """
    there is no good way yet to statically assert a class is a "real" dataclass, instead we use this best-effort protocol
    to type check if a class is a dataclass cf. https://stackoverflow.com/a/55240861/7723404
    """

    __dataclass_fields__: dict[str, Any]


class JsonSnakeCaseSerializableMixin(DataclassLike):
    def to_dict(
        self,
        encode_json: bool = False,
        exclude_none: bool = False,
        exclude_defaults: bool = False,
    ) -> dict:
        #  we need to runtime check that we are actually a dataclass here since we can't stop people from inheriting
        # from the mixin without being a dataclass AND statically type-narrow to fulfill the static typing requirements
        if not is_dataclass(self) or not isinstance(self, DataclassLike):
            raise TypeError("JSONSerialization is only supported for dataclasses")

        return apischema.serialize(
            aliaser=apischema.utils.to_snake_case,
            additional_properties=True,
            obj=self,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
        )

    @classmethod
    def from_dict(cls: type[T], kvs: dict[str, Any], *, infer_missing: bool = False) -> T:
        # we need to runtime check that we are actually a dataclass here since we can't stop people from inheriting
        # from the mixin without being a dataclass
        if not is_dataclass(cls):
            raise TypeError("JSONDeserialization is only supported for dataclasses")

        return cast(
            T,
            apischema.deserialize(
                aliaser=apischema.utils.to_snake_case,
                additional_properties=True,
                data=kvs,
                type=cls,
            ),
        )



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass, field
from typing import Mapping, Optional, TypeVar, Union

# noinspection PyProtectedMember
from uai_openlabel.elements.action import ActionInFrame

# noinspection PyProtectedMember
from uai_openlabel.elements.context import ContextInFrame

# noinspection PyProtectedMember
from uai_openlabel.elements.event import EventInFrame

# noinspection PyProtectedMember
from uai_openlabel.elements.object import ObjectInFrame

# noinspection PyProtectedMember
from uai_openlabel.elements.relation import Relation

# noinspection PyProtectedMember
from uai_openlabel.serializer import JsonSnakeCaseSerializableMixin

# noinspection PyProtectedMember
from uai_openlabel.stream.stream import Stream

# noinspection PyProtectedMember
from uai_openlabel.transform import Transform

# noinspection PyProtectedMember
from uai_openlabel.types_and_constants import (
    ActionUid,
    ContextUid,
    EventUid,
    Number,
    ObjectUid,
    RelationUid,
    StreamUid,
)

__all__: list[str] = []


@dataclass
class FrameProperties(JsonSnakeCaseSerializableMixin):
    timestamp: Optional[Union[str, Number]] = field(default=None)
    streams: Optional[Mapping[StreamUid, Stream]] = field(default=None)
    transforms: Optional[Mapping[str, Transform]] = field(default=None)


T = TypeVar("T", bound="Frame")


@dataclass
class Frame(JsonSnakeCaseSerializableMixin):
    """
    In case time-information is needed, for example, for labeling video sequences, the item frames contains a dictionary of containers at frame level.

    The structure of Frame is similar to the openlabel value as it contains dictionaries for the elements,
    meaning objects, actions, events, contexts, and relations.

    Only the dynamic information inside them is detailed.
    In addition, frame_properties may contain information about timestamping details,
    or transforms of specific coordinate systems and other stream properties.
    """

    actions: Optional[Mapping[ActionUid, ActionInFrame]] = field(default=None)
    contexts: Optional[Mapping[ContextUid, ContextInFrame]] = field(default=None)
    events: Optional[Mapping[EventUid, EventInFrame]] = field(default=None)
    frame_properties: Optional[FrameProperties] = field(default=None)
    objects: Optional[Mapping[ObjectUid, ObjectInFrame]] = field(default=None)
    relations: Optional[Mapping[RelationUid, Relation]] = field(
        default=None
    )  # The spec doesn't specify this correctly, see 7.5. Is there sth like a RelationInFrame we should use here?

    @classmethod
    def example(
        cls: type[T],
        timestamp: str,
        transform_from_and_to: tuple[str, str],
        transform_translation: tuple[float, float, float],
        frame_objects: dict[ObjectUid, ObjectInFrame],
    ) -> T:
        src, dst = transform_from_and_to
        return cls(
            actions=None,
            contexts=None,
            events=None,
            frame_properties=FrameProperties(
                timestamp=timestamp,
                transforms={
                    f"{src}_to_{dst}": Transform.no_rotation_example(src=src, dst=dst, translation=transform_translation)
                },
            ),
            objects=frame_objects,
            relations=None,
        )



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import builtins
from dataclasses import dataclass, field
from typing import Mapping, Optional, Sequence, TypeVar, Union

from apischema.metadata import required

# noinspection PyProtectedMember
from uai_openlabel.data_types.generic_data import Attributes

# noinspection PyProtectedMember
from uai_openlabel.data_types.geometric_data import (
    Poly2D,
    Poly3D,
    RotatedTwoDBoundingBox,
    ThreeDBoundingBoxEuler,
    ThreeDBoundingBoxQuaternion,
    TwoDBoundingBox,
)

# noinspection PyProtectedMember
from uai_openlabel.elements.element_data_pointer import ElementDataPointer

# noinspection PyProtectedMember
from uai_openlabel.frame_interval import FrameInterval, no_default

# noinspection PyProtectedMember
from uai_openlabel.serializer import JsonSnakeCaseSerializableMixin

# noinspection PyProtectedMember
from uai_openlabel.types_and_constants import AttributeName, OntologyUid, ResourceUid

__all__: list[str] = []


D = TypeVar("D", bound="ObjectData")


@dataclass
class ObjectData(Attributes):
    # TODO When implementing these, don't forget to add them to data_to_pointer_type_mapping.py
    # boolean, num, text, vec defined in Attributes
    # area_reference: Optional[Sequence[...]] = field(default=None)
    bbox: Optional[Sequence[TwoDBoundingBox]] = field(default=None)
    # binary: Optional[Sequence[...]] = field(default=None)
    cuboid: Optional[Sequence[Union[ThreeDBoundingBoxEuler, ThreeDBoundingBoxQuaternion]]] = field(default=None)
    # image: Optional[Sequence[...]] = field(default=None)
    # line_reference: Optional[Sequence[...]] = field(default=None)
    # mat: Optional[Sequence[...]] = field(default=None)
    # mesh: Optional[Sequence[...]] = field(default=None)
    # point2d: Optional[Sequence[...]] = field(default=None)
    # point3d: Optional[Sequence[...]] = field(default=None)
    poly2d: Optional[Sequence[Poly2D]] = field(default=None)
    poly3d: Optional[Sequence[Poly3D]] = field(default=None)
    rbbox: Optional[Sequence[RotatedTwoDBoundingBox]] = field(default=None)

    @classmethod
    def cuboid_with_dynamic_attributes_example(
        cls: type[D],
        toggle_attribute_value: bool,
        cuboid_translation: tuple[float, float, float],
    ) -> D:
        example = cls.dynamic_attributes_example(toggle_attribute_value)

        euler_angles = (0.0, 0.0, 0.0)
        cuboid_size = (4.1, 1.75, 1.45)
        example.cuboid = [
            ThreeDBoundingBoxEuler(
                val=cuboid_translation + euler_angles + cuboid_size,
                name="bounding_box",
            )
        ]
        return example


T = TypeVar("T", bound="Object")


@dataclass
class Object(JsonSnakeCaseSerializableMixin):
    """
    A structure to represent information about physical entities in scenes.
    Examples of objects are pedestrians, cars, the ego-vehicle, traffic signs, lane markings, building, and trees.

    :param name: A friendly identifier of the element, is not unique but employed by human users to rapidly identify elements in the scene, for example, Peter.
    :param type: The classification of the object
    :param frame_intervals: An array of frame intervals where the object exists.
    :param ontology_uid: A string identifier of the ontology which contains the definition of the type of the element.
    :param resource_uid: A string identifier of the resource which contains additional information on the type of the element.
    :param object_data: Container of static information about the object.
    :param object_data_pointers: Pointers to element data at frames.
    """

    name: str = field(default_factory=lambda: no_default(field="Object.name"), metadata=required)
    type: str = field(default_factory=lambda: no_default(field="Object.type"), metadata=required)

    frame_intervals: Optional[Sequence[FrameInterval]] = field(default=None)
    ontology_uid: Optional[OntologyUid] = field(default=None)
    resource_uid: Optional[ResourceUid] = field(default=None)
    object_data: Optional[ObjectData] = field(default=None)
    object_data_pointers: Optional[Mapping[AttributeName, ElementDataPointer]] = field(default=None)

    @classmethod
    def car_example(
        cls: builtins.type[T],
        name: str,
        frame_interval: FrameInterval,
        object_data_pointers: Optional[dict[AttributeName, ElementDataPointer]] = None,
    ) -> T:
        return cls(
            name=name,
            type="car",
            frame_intervals=[frame_interval],
            object_data=ObjectData.static_attributes_example(),
            object_data_pointers=object_data_pointers,
        )


F = TypeVar("F", bound="ObjectInFrame")


@dataclass
class ObjectInFrame(JsonSnakeCaseSerializableMixin):
    object_data: ObjectData = field(
        default_factory=lambda: no_default(field="ObjectInFrame.object_data"),
        metadata=required,
    )

    @classmethod
    def example(
        cls: builtins.type[F],
        toggle_attribute_values: bool,
        cuboid_translation: tuple[float, float, float],
    ) -> F:
        return cls(
            object_data=ObjectData.cuboid_with_dynamic_attributes_example(
                toggle_attribute_value=toggle_attribute_values,
                cuboid_translation=cuboid_translation,
            )
        )



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass, field
from typing import Mapping, Optional, Sequence

from apischema.metadata import required

# noinspection PyProtectedMember
from uai_openlabel.data_types.generic_data import Attributes

# noinspection PyProtectedMember
from uai_openlabel.elements.element_data_pointer import ElementDataPointer

# noinspection PyProtectedMember
from uai_openlabel.frame_interval import FrameInterval, no_default

# noinspection PyProtectedMember
from uai_openlabel.serializer import JsonSnakeCaseSerializableMixin

# noinspection PyProtectedMember
from uai_openlabel.types_and_constants import AttributeName, OntologyUid, ResourceUid

__all__: list[str] = []


@dataclass
class Action(JsonSnakeCaseSerializableMixin):
    """
    A description of semantically meaningful acts being done.
    They may be defined for several frame intervals, similar to objects, for example, isWalking.
    """

    name: str = field(default_factory=lambda: no_default(field="Action.name"), metadata=required)
    type: str = field(default_factory=lambda: no_default(field="Action.type"), metadata=required)

    frame_intervals: Optional[Sequence[FrameInterval]] = field(default=None)
    ontology_uid: Optional[OntologyUid] = field(default=None)
    resource_uid: Optional[ResourceUid] = field(default=None)
    action_data: Optional[Attributes] = field(default=None)
    action_data_pointers: Optional[Mapping[AttributeName, ElementDataPointer]] = field(default=None)


@dataclass
class ActionInFrame(JsonSnakeCaseSerializableMixin):
    action_data: Attributes = field(
        default_factory=lambda: no_default(field="ActionInFrame.action_data"),
        metadata=required,
    )



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import logging
from typing import Any, Optional, Sequence, TypeVar, Union

__all__: list[str] = []


logger = logging.getLogger(__name__)


def no_default(field: str) -> Any:
    message = f"Must set a value for {field}"
    raise ValueError(message)


T = TypeVar("T")
V = TypeVar("V")


def unpack_sequence_of_length_1(
    seq: Union[T, list[T], tuple[T]],
    field_name_for_logging: str,
    stand_in_for_empty_seq: Optional[T] = None,
) -> T:
    if not isinstance(seq, list) and not isinstance(seq, tuple):
        return seq

    if stand_in_for_empty_seq is not None and len(seq) == 0:
        return stand_in_for_empty_seq

    if len(seq) != 1:
        raise ValueError(
            f"Got a sequence of values when initializing {field_name_for_logging}. "
            "Only the special case of a sequence length of 1 can be handled. "
        )
    return seq[0]


def convert_values(
    values: Sequence[Any],
    conversion_target: type[T],
    dont_convert: Sequence[type[V]],
    field_name_for_logging: str,
) -> Sequence[Union[T, V]]:
    """No need to add conversion target to dont_convert"""
    extended_dont_convert = [conversion_target] + [t for t in dont_convert]
    to_be_converted = [v.__class__ not in extended_dont_convert for v in values]
    if any(to_be_converted):
        logger.info(
            "The values of field %s aren't of types %s and will be converted to %s",
            field_name_for_logging,
            [t.__name__ for t in extended_dont_convert],
            conversion_target.__name__,
        )

    # see https://github.com/python/mypy/issues/10343
    converted_val = [conversion_target(v) if to_be_converted[i] else v for i, v in enumerate(values)]  # type: ignore[call-arg]
    return converted_val



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass, field
from typing import Literal, Optional, TypeVar

# noinspection PyProtectedMember
from uai_openlabel.serializer import JsonSnakeCaseSerializableMixin

__all__: list[str] = []


T = TypeVar("T", bound="Metadata")


@dataclass
class Metadata(JsonSnakeCaseSerializableMixin):
    schema_version: Literal["1.0.0"] = field(default="1.0.0")
    """	Version number of the OpenLABEL schema this annotation JSON object follows."""

    annotator: Optional[str] = field(default=None)
    """Name or description of the annotator that created the annotations."""

    comment: Optional[str] = field(default=None)
    """Additional information or description about the annotation content."""

    file_version: Optional[str] = field(default=None)
    """Version number of the OpenLABEL annotation content."""

    name: Optional[str] = field(default=None)
    """Name of the OpenLABEL annotation content."""

    tagged_file: Optional[str] = field(default=None)
    """File name or URI of the data file being tagged."""

    @classmethod
    def example(cls: type[T]) -> T:
        return cls(
            schema_version="1.0.0",
            annotator="understand.AI GmbH",
            comment="This is an example OpenLabel file. All labels are imaginary.",
            file_version="1",
            name="Example export",
        )



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import builtins
from dataclasses import dataclass, field
from enum import Enum
from typing import Iterable, Iterator, Optional, Sequence, TypeVar, Union

from apischema.metadata import required

# noinspection PyProtectedMember
from uai_openlabel.serializer import JsonSnakeCaseSerializableMixin

# noinspection PyProtectedMember
from uai_openlabel.types_and_constants import (
    AttributeName,
    CoordinateSystemUid,
    Number,
    ObjectUid,
)

# noinspection PyProtectedMember
from uai_openlabel.utils import convert_values, no_default, unpack_sequence_of_length_1

__all__: list[str] = []


class BooleanType(Enum):
    Value = "value"


B = TypeVar("B", bound="BooleanData")


@dataclass
class BooleanData(JsonSnakeCaseSerializableMixin):
    val: bool = field(default_factory=lambda: no_default(field="BooleanData.val"), metadata=required)

    attributes: Optional["Attributes"] = field(default=None)
    coordinate_system: Optional[CoordinateSystemUid] = field(default=None)
    name: Optional[AttributeName] = field(default=None)
    type: Optional[BooleanType] = field(default=None)

    def __post_init__(self) -> None:
        """apischema doesn't allow subclasses of the annotated types, so we need to cast them."""
        field_name_for_logging = f"{self.__class__.__name__}.val"
        self.val = unpack_sequence_of_length_1(self.val, field_name_for_logging)
        converted_val = convert_values(
            values=[self.val],
            conversion_target=bool,
            dont_convert=[],
            field_name_for_logging=field_name_for_logging,
        )[0]
        self.val = converted_val

    @classmethod
    def static_example(cls: builtins.type[B]) -> B:
        """An attribute that only has one value."""
        return cls(
            val=True,
            name="has_driver",
            type=BooleanType.Value,
        )

    @classmethod
    def dynamic_example(cls: builtins.type[B], toggle_value: bool = False) -> B:
        """An attribute that may take one of two values, toggleable via the toggle_value switch."""
        return cls(
            val=False if toggle_value else True,
            name="brake_lights_on",
            type=BooleanType.Value,
        )


class NumberType(Enum):
    Value = "value"
    Min = "min"
    Max = "max"


N = TypeVar("N", bound="NumberData")


@dataclass
class NumberData(JsonSnakeCaseSerializableMixin):
    val: Number = field(default_factory=lambda: no_default(field="NumberData.val"), metadata=required)

    attributes: Optional["Attributes"] = field(default=None)
    coordinate_system: Optional[CoordinateSystemUid] = field(default=None)
    name: Optional[AttributeName] = field(default=None)
    type: Optional[NumberType] = field(default=None)

    def __post_init__(self) -> None:
        """apischema doesn't allow subclasses of the annotated types, so we need to cast them."""
        field_name_for_logging = f"{self.__class__.__name__}.val"
        self.val = unpack_sequence_of_length_1(self.val, field_name_for_logging)
        converted_val = convert_values(
            values=[self.val],
            conversion_target=float,
            dont_convert=[int],
            field_name_for_logging=field_name_for_logging,
        )[0]
        self.val = converted_val

    @classmethod
    def static_example(cls: builtins.type[N]) -> N:
        """An attribute that only has one value."""
        return cls(
            val=4,
            name="nr_wheels",
            type=NumberType.Value,
        )

    @classmethod
    def dynamic_example(cls: builtins.type[N], toggle_value: bool = False) -> N:
        """An attribute that may take one of two values, toggleable via the toggle_value switch."""
        return cls(
            val=5 if toggle_value else 4,
            name="leaves_on_hood",
            type=NumberType.Value,
        )


class TextType(Enum):
    Value = "value"


T = TypeVar("T", bound="TextData")


@dataclass
class TextData(JsonSnakeCaseSerializableMixin):
    val: str = field(default_factory=lambda: no_default(field="TextData.val"), metadata=required)

    attributes: Optional["Attributes"] = field(default=None)
    coordinate_system: Optional[CoordinateSystemUid] = field(default=None)
    name: Optional[AttributeName] = field(default=None)
    type: Optional[TextType] = field(default=None)

    def __post_init__(self) -> None:
        """apischema doesn't allow subclasses of the annotated types, so we need to cast them."""
        field_name_for_logging = f"{self.__class__.__name__}.val"
        self.val = unpack_sequence_of_length_1(self.val, field_name_for_logging, "")
        converted_val = convert_values(
            values=[self.val],
            conversion_target=str,
            dont_convert=[ObjectUid],
            field_name_for_logging=field_name_for_logging,
        )[0]
        self.val = converted_val

    @classmethod
    def static_example(cls: builtins.type[T]) -> T:
        """An attribute that only has one value."""
        return cls(
            val="red",
            name="color",
            type=TextType.Value,
        )

    @classmethod
    def dynamic_example(cls: builtins.type[T], toggle_value: bool = False) -> T:
        """An attribute that may take one of two values, toggleable via the toggle_value switch."""
        return cls(
            val="NE" if toggle_value else "ENE",
            name="compass_heading",
            type=TextType.Value,
        )


class VectorType(Enum):
    Values = "values"
    Ranges = "ranges"


V = TypeVar("V", bound="VectorData")


@dataclass
class VectorData(JsonSnakeCaseSerializableMixin):
    val: Union[Sequence[Number], Sequence[str]] = field(
        default_factory=lambda: no_default(field="VectorData.val"), metadata=required
    )

    attributes: Optional["Attributes"] = field(default=None)
    coordinate_system: Optional[CoordinateSystemUid] = field(default=None)
    name: Optional[AttributeName] = field(default=None)
    type: Optional[VectorType] = field(default=None)

    def __post_init__(self) -> None:
        """apischema doesn't allow subclasses of the annotated types, so we need to cast them."""
        field_name_for_logging = f"{self.__class__.__name__}.val"
        converted_val = convert_values(
            values=self.val,
            conversion_target=float,
            dont_convert=[int, str, ObjectUid],
            field_name_for_logging=field_name_for_logging,
        )
        self.val = tuple(converted_val)

    @classmethod
    def static_example(cls: builtins.type[V]) -> V:
        """An attribute that only has one value."""
        return cls(
            val=["tobi", "veith", "daniel"],
            name="previous_drivers",
            type=VectorType.Values,
        )

    @classmethod
    def dynamic_example(cls: builtins.type[V], toggle_value: bool = False) -> V:
        """An attribute that may take one of two values, toggleable via the toggle_value switch."""
        return cls(
            val=[0.2, 0.7, 0.3] if toggle_value else [0.5, 0.1, 0.4],
            name="radio_coefficients",
            type=VectorType.Values,
        )


GenericData = Union[BooleanData, NumberData, TextData, VectorData]


A = TypeVar("A", bound="Attributes")


@dataclass
class Attributes(JsonSnakeCaseSerializableMixin, Iterable[GenericData]):
    boolean: Optional[Sequence[BooleanData]] = field(default=None)
    num: Optional[Sequence[NumberData]] = field(default=None)
    text: Optional[Sequence[TextData]] = field(default=None)
    vec: Optional[Sequence[VectorData]] = field(default=None)

    def __iter__(self) -> Iterator[GenericData]:
        """Iterates over all the data, in all fields, in this order: boolean, num, text, vec."""
        for field_name in self.__dataclass_fields__:
            attributes = self.__getattribute__(field_name)
            if attributes is None:
                continue
            for a in attributes:
                yield a

    @classmethod
    def static_attributes_example(cls: builtins.type[A]) -> A:
        """Contains attributes meant to be static."""
        return cls(
            boolean=[BooleanData.static_example()],
            num=[NumberData.static_example()],
            text=[TextData.static_example()],
            vec=[VectorData.static_example()],
        )

    @classmethod
    def dynamic_attributes_example(cls: builtins.type[A], toggle_value: bool = False) -> A:
        """Contains attributes meant to be dynamic. Use the toggle_value switch to vary them."""
        return cls(
            boolean=[BooleanData.dynamic_example(toggle_value)],
            num=[NumberData.dynamic_example(toggle_value)],
            text=[TextData.dynamic_example(toggle_value)],
            vec=[VectorData.dynamic_example(toggle_value)],
        )



# Copyright © 2024 understandAI GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass, field
from typing import Optional, Union

# noinspection PyProtectedMember
from uai_openlabel.serializer import JsonSnakeCaseSerializableMixin

# noinspection PyProtectedMember
from uai_openlabel.stream.camera_intrinsics import (
    CustomCameraIntrinsics,
    FisheyeCameraIntrinsics,
    PinholeCameraIntrinsics,
)

# noinspection PyProtectedMember
from uai_openlabel.types_and_constants import FrameUid, Number

__all__: list[str] = []


@dataclass
class SyncByFrameStream(JsonSnakeCaseSerializableMixin):
    """
    frame_stream: This is the internal frame number inside the stream this OpenLABEL frame corresponds to.
    timestamp: The timestamp of the data source in this particular Stream object
    """

    frame_stream: Optional[FrameUid] = field(default=None)
    timestamp: Optional[Union[str, Number]] = field(default=None)


@dataclass
class SyncByFrameShift(JsonSnakeCaseSerializableMixin):
    """
    frame_shift: Fixed shift or difference between the OpenLABEL master frame count and this stream's internal frame count.
    """

    frame_shift: Optional[int] = field(default=None)


@dataclass
class PinholeCameraStreamProperties(JsonSnakeCaseSerializableMixin):
    sync: Optional[Union[SyncByFrameStream, SyncByFrameShift]] = field(default=None)
    intrinsics_pinhole: Optional[PinholeCameraIntrinsics] = field(default=None)


@dataclass
class FisheyeCameraStreamProperties(JsonSnakeCaseSerializableMixin):
    sync: Optional[Union[SyncByFrameStream, SyncByFrameShift]] = field(default=None)
    intrinsics_fisheye: Optional[FisheyeCameraIntrinsics] = field(default=None)


@dataclass
class CustomCameraStreamProperties(JsonSnakeCaseSerializableMixin):
    sync: Optional[Union[SyncByFrameStream, SyncByFrameShift]] = field(default=None)
    intrinsics_custom: Optional[CustomCameraIntrinsics] = field(default=None)


StreamProperties = Union[
    PinholeCameraStreamProperties,
    FisheyeCameraStreamProperties,
    CustomCameraStreamProperties,
]
